// cvwlle.cc -- A constrained-variable weighted log-likelihood optimizer
//              with regularizer factors set by cross-validation
//
// Mark Johnson, 1st March 2004
//
// This is a version of wlle.cc reworked to use the tao-optimizer.h
// routines (so most of the Petsc/Tao junk is now moved out of this
// file).  It also sets regularizer factors by cross-validation.
// It writes out the weights file after each cross-validation estimation.
//
// This optimizier uses the TAO constrained optimization method tao_blmvm
// to avoid the variable discontinuity at zero if the -cv flag is set.
//
// For each variable x[i] in the original problem we introduce a pair of
// variables xp[i] and xn[i] and a pair of constraints xp[i] >= 0, xn[i] <= 0.
// Then x[i] = xp[i] + xn[i].
//
// The regularized function Q that is optimized is:
//
//  Q(xp,xn) = s * ( L(xp+xn) + R(xp,xn) )
//
// where s is a user-specified parameter, L is the unregularized loss 
// function and R is the regularizer:
//
//  R(xp,xn) = c * sum_i (pow(fabs(xp[i]),p) + pow(fabs(xn[i]),p))
//
// where c and p are user-specifier adjustable parameters.  (The fabs()
// is there just in case the optimization routine temporarily proposes an
// infeasible solution).
//
// Note that
//
//  d Q/d xp[i] = s * d L/ d xp[i] + s * c * p * sum_i pow(fabs(xp[i]),p-1) * sign(xp[i])
//
// where sign(xp[i]) is +1 if xp[i] is non-negative and -1 otherwise.

const char usage[] =
"cvwlle version of 28th February 2004\n"
"\n"
"	A constrained-variable weighted log-likelihood optimizer\n"
"	with regularizer factors set by cross-validation.\n"
"\n"
"cvwlle estimates feature weights that estimate the parameters of a\n"
"log-linear model by minimizing the regularized log loss of the feature\n"
"weights using the LVLM or BLVLM optimizer from the Petsc/Tao\n"
"optimization package (see http://www-fp.mcs.anl.gov/tao/ for details).\n"
"It can deal with partially labeled data (i.e., training instances\n"
"consist of one or more \"winners\" and one or more \"losers\").\n"
"\n"
"Usage: cvwlle [-help] [-debug debug_level] [-c0 c0] [-ci (...)] [-p p] [-r r] [-s s] [-cv]\n"
"              [-Pyx_factor f] [-Px_propto_g] \n"
"              [-o weights-file]  [-e eval-file]\n"
"              [-fcb base] [-ns ns] [-f feat-file]\n"
"              tao-options*\n"
"	       < train-file\n"
"\n"
"where:\n"
"\n"
" debug_level > 0 controls the amount of output produced\n"
"\n"
" data respectively (best if slight over-estimates, ignored if the data\n"
" specifies NS),\n"
"\n"
" train-file and eval-file are files from which training and evaluation\n"
" data are read (if eval-file ends in the suffix .bz2 then bzcat is used\n"
" to read it; if no eval-file is specified, then the program tests on the\n"
" training data),\n"
"\n"
" weights-file is a file to which the estimated weights are written,\n"
"\n"
" feat-file is a file of <featclass> <featuredetails> lines, used for\n"
" cross-validating regularizer weights,\n"
"\n"
" ns is the maximum number of ':' characters in a <featclass>, used to\n"
" determine how features are binned into feature classes (ns = -1 bins\n"
" all features into the same class)\n"
"\n"
" -fcb specifies that features should be binned by their counts log base\n"
"   (this prefixes the feature class with the floor of the log base)\n"
"\n"
" r specifies that the weights are initialized to random values in\n"
"   [-r ... +r],\n"
"\n"
" -Pyx_factor f indicates that a parse should be taken as correct\n"
"   proportional to f raised to its f-score, and\n"
"\n"
" -Px_propto_g indicates that each sentence is weighted by the number of\n"
"   edges in its gold parse.\n"
"\n"
"The function that the program minimizes is:\n"
"\n"
"   Q(w) = s * (- L(w) + c * sum_j pow(fabs(w[j]), p) ), where:\n"
"\n"
"   L(w) = log conditional likelihood of the training data\n"
"\n"
"        = sum_i P~(x_i) ( E_P~[w.f|x_i] - log Z_w(x_i) ), and\n"
"\n"
"   E_P~[w.f|x] = sum_y w . f(x,y) P~(y|x)\n"
"\n"
"   Z_w(x) = sum_y exp(w . f(x,y))\n"
"\n"
"   P(y|x) = exp w . f(x,y) / Z_w(x)\n"
"\n"
"   P~(x) and P~(y|x) is the empirical (training) distribution.\n"
"\n"
"The -cv option instructs the program to optimize a function defined in\n"
"terms of vectors of variables u[] and v[], where w[j] = u[j] + v[j]\n"
"and v[j] <= 0 <= u[j], otherwise it optimizes the w[j] directly.\n"
"\n"
"With debug = 0, the program writes a single line to stdout:\n"
"\n"
"c p r s it nzeroweights/nweights neglogP/nsentence ncorrect/nsentences\n"
"\n"
"With debug >= 10, the program writes out a histogram of weights as well\n"
"\n"
"Data format:\n"
"-----------\n"
"\n"
"<Data>     --> [S=<NS>] <Sentence>*\n"
"<Sentence> --> [G=<G>] N=<N> <Parse>*\n"
"<Parse>    --> [P=<P>] [W=<W>] <FC>*,\n"
"<FC>       --> <F>[=<C>]\n"
"\n"
"NS is the number of sentences.\n"
"\n"
"Each <Sentence> consists of N <Parse>s.  <G> is the gold standard\n"
"score.  To get parsing precision and recall results, set <G> to the\n"
"number of edges in the gold standard parse.  To get accuracy results,\n"
"set <G> to 1 (the default).\n"
"\n"
"A <Parse> consists of <FC> pairs.  <P> is the parse's possible highest\n"
"score and <W> is the parse's actual score.  To get parsing precision and\n"
"recall results, set <P> to the number of edges in the parse and <W> to\n"
"the number of edges in common between the gold and parse trees.\n"
"\n"
"A <FC> consists of a feature (a non-negative integer) and an optional\n"
"count (a real).\n"
"\n"
"The default for all numbers except <W> is 1.  The default for <W> is 0.\n";

#define _GLIBCPP_CONCEPT_CHECKS  // uncomment this for checking

#include <boost/lexical_cast.hpp>
#include <cassert>
#include <cctype>
#include <cerrno>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <vector>

#include "data.h"
#include "powell.h"
#include "utility.h"
#include "tao-optimizer.h"

typedef std::vector<double> doubles;
typedef std::vector<size_t> size_ts;

int debug_level = 0;

void print_histogram(int nx, const double x[], int nbins=20) {
  int nx_nonzero = 0;
  for (int i = 0; i < nx; ++i)
    if (x[i] != 0)
      ++nx_nonzero;

  std::cout << "#   There are " << nx_nonzero << " non-zero values and " 
	    << nx-nx_nonzero << " zero values." << std::endl;

  if (nx_nonzero > 0) {
    std::vector<double> s;
    s.reserve(nx_nonzero);
    for (int i = 0; i < nx; ++i)
      if (x[i] != 0)
	s.push_back(x[i]);
    std::sort(s.begin(), s.end());
    for (int i = 0; i <= nbins; ++i) {
      int j = i*(nx_nonzero-1);
      j /= nbins;
      std::cout << float(i)/float(nbins) << '\t' << s[j] << std::endl;
    }
  }
}  // print_histogram()

// f_df() evaluates the statistics of the corpus, and prints the f-score
//  if required
//
double f_df(corpus_type* corpus, double x[], double df_dx[]) {
  Float sum_g = 0, sum_p = 0, sum_w = 0;
  Float L = corpus_stats(corpus, &x[0], &df_dx[0], &sum_g, &sum_p, &sum_w);
  
  if (debug_level >= 1000)
    std::cout << "f score = " << 2*sum_w/(sum_g+sum_p) << ", " << std::flush;
  
  assert(finite(L));
  return L;
}


// Unconstrained is for unconstrained optimization
//
struct Unconstrained {
  corpus_type* corpus;
  const size_ts& f_c;	//!< feature -> cross-validation class
  const doubles& cs;	//!< cross-validation class -> cross-validation "constant"
  double p, s;
  int it;

  Unconstrained(corpus_type* corpus, const size_ts& f_c, 
		const doubles& cs, double p, double s) 
    : corpus(corpus), f_c(f_c), cs(cs), p(p), s(s), it(0) 
  { 
    assert(f_c.size() == corpus->nfeatures);
    for (size_t f = 0; f < f_c.size(); ++f)
      assert(f_c[f] < cs.size());
  }

  double operator() (int nx, double x[], double df_dx[]) {
    it++;
    if (debug_level >= 1000)      
      std::cout << "it = " << it << ", " << std::flush;

    Float L = f_df(corpus, x, df_dx);

    if (s != 1) {
      L *= s;

      for (int i = 0; i < nx; ++i) {
	assert(finite(df_dx[i]));
	df_dx[i] *= s;
      }
    }

    assert(size_t(nx) == corpus->nfeatures);
    assert(f_c.size() == size_t(nx));

    double R = 0;
    for (int i = 0; i < nx; ++i) 
      R +=  cs[f_c[i]] * pow(fabs(x[i]), p);
    R *= s;

    double Q = L + R;
    
    if (debug_level >= 1000)
      std::cout << "Q = " << Q << " = L = " << L << " + R = " << R << std::endl;

    assert(finite(Q));
    
    double sp = s * p;
    for (int i = 0; i < nx; ++i) 
      df_dx[i] += sp * cs[f_c[i]] * pow(fabs(x[i]), p-1) 
	* (x[i] >= 0 ? (x[i] == 0 ? 0 : 1) : -1);

    if (debug_level >= 10000) {
      std::cout << "Histogram of derivatives:" << std::endl;
      print_histogram(nx, &df_dx[0]);
      std::cout << "--------------------------" << std::endl;
    }

    for (int i = 0; i < nx; ++i)
      assert(finite(df_dx[i]));
    return Q;
  }   // Unconstrained::operator()

};  // Unconstrained{}


// Constrained{} for the constrained optimization.
//
struct Constrained {
  corpus_type* corpus;
  const size_ts& f_c;	//!< feature -> cross-validation class
  const doubles& cs;	//!< cross-validation class -> cross-validation "constant"
  double p, s;
  int it;
  doubles x, df_dx;

  Constrained(corpus_type* corpus, const size_ts& f_c, 
	      const doubles& cs, double p, double s) 
    : corpus(corpus), f_c(f_c), cs(cs), p(p), s(s), it(0),
      x(corpus->nfeatures, 0), df_dx(corpus->nfeatures) { 
    assert(f_c.size() == corpus->nfeatures);
    for (size_t f = 0; f < f_c.size(); ++f)
      assert(f_c[f] < cs.size());
  }

  double operator() (int nx2, double x2[], double df_dx2[]) {
    it++;
    if (debug_level >= 1000)      std::cout << "it = " << it << ", " << std::flush;
    
    size_t nx = corpus->nfeatures;
    assert(nx2 = 2 * nx);
    assert(x.size() == nx);
    for (size_t i = 0; i < nx; ++i) {
      assert(finite(x2[i]));
      assert(finite(x2[i+nx]));
      x[i] = x2[i] + x2[i+nx];
      assert(finite(x[i]));
    }

    Float L = f_df(corpus, &x[0], &df_dx[0]);
    
    L *= s;

    for (size_t i = 0; i < nx; ++i) {
      assert(finite(df_dx[i]));
      df_dx2[i] = df_dx2[i+nx] = s * df_dx[i];
    }

    double R = 0;
    for (size_t i = 0; i < nx; ++i)
      R += cs[f_c[i]] * pow(fabs(x2[i]), p) + pow(fabs(x2[i+nx]), p);
    R *= s;

    double Q = L + R;
    
    if (debug_level >= 1000)
      std::cout << "Q = " << Q << " = L = " << L << " + R = " << R << std::endl;

    assert(finite(Q));
    
    double sp = s * p;
    for (size_t i = 0; i < nx; ++i) {
      df_dx2[i] += sp * cs[f_c[i]] * pow(fabs(x2[i]), p-1) * (x2[i] >= 0 ? 1 : -1);
      df_dx2[i+nx] += sp * cs[f_c[i]] * pow(fabs(x2[i+nx]), p-1) * (x2[i+nx] > 0 ? 1 : -1);
    }

    if (debug_level >= 10000) {
      std::cout << "Histogram of derivatives:" << std::endl;
      print_histogram(nx, &df_dx[0]);
      std::cout << "--------------------------" << std::endl;
    }

    for (int i = 0; i < nx2; ++i)
      assert(finite(df_dx2[i]));
    return Q;
  }   // Constrained::operator()

};  // Constrained{}


// The Estimator1 does one round of estimation
//
struct Estimator1 {
  typedef std::vector<size_t> size_ts;
  typedef std::vector<double> doubles;
  
  corpus_type* train;	//!< training data
  size_t nx;		//!< number of features
  corpus_type* eval;	//!< evaluation data
  double c0;		//!< default regularizer factor
  double p;		//!< regularizer power
  double r;		//!< random initialization
  double s;		//!< scale factor
  bool cv;		//!< use constrained variables
  bool opt_fscore;	//!< optimize f-score or - log likelihood

  doubles x;		//!< feature -> weight
  doubles x0;		//!< feature -> weight; initial starting point for later optimization
  size_ts f_c;		//!< feature -> cross-validation class
  doubles cs;		//!< cross-validation class -> factor
  size_t nc;		//!< number of cross-validation classes
  size_t nits;		//!< number of iterations of last round
  size_t sum_nits;	//!< total number of iterations
  size_t nrounds;	//!< number of cross-validation rounds so far
  std::string weightsfile; //!< name of weights file

  typedef std::map<std::string,size_t> S_C;
  S_C identifier_regclass; //!< map from feature class identifiers to regularization class
  typedef std::vector<std::string> Ss;
  Ss regclass_identifiers; //!< vector of class identifiers

  Estimator1(corpus_type* train, corpus_type* eval, 
	     double c0, double p, double r, double s, bool cv, 
	     bool opt_fscore = false, const char* weightsfile = NULL) 
    : train(train), nx(train->nfeatures), eval(eval), 
      c0(c0), p(p), r(r), s(s), cv(cv), 
      opt_fscore(opt_fscore), x(nx), x0(nx), f_c(nx), cs(1, c0), nc(1), 
      nits(0), sum_nits(0), nrounds(0), 
      weightsfile(weightsfile == NULL ? "" : weightsfile)
  { }  // Estimator1::Estimator1()

  // optimize1() performs a single round of estimation
  //
  size_t optimize1(const doubles& ccs, doubles& xx) {
    size_t nits = 0;
    assert(xx.size() == nx);
    assert(x0.size() == nx);
    if (cv) {

      // Constrained variable optimization

      Constrained fn(train, f_c, ccs, p, s);
      tao_constrained_optimizer<Constrained> tao_opt(2*nx, fn);

      if (r != 0) 
	for (size_t i = 0; i < 2*nx; ++i)
	  tao_opt[i] = r*double(random()-RAND_MAX/2)/double(RAND_MAX/2);
    
      for (size_t i = 0; i < nx; ++i) {
	tao_opt.lower_bound[i] = 0;
	tao_opt.lower_bound[i+nx] = TAO_NINFINITY;
	tao_opt.upper_bound[i] = TAO_INFINITY;
	tao_opt.upper_bound[i+nx] = 0;

	// copy the default starting point into tao_opt[]

	if (x0[i] > 0)
	  tao_opt[i] = x0[i];
	else
	  tao_opt[i+nx] = x0[i];
      }
      
      tao_opt.optimize();
    
      nits = fn.it;
      
      for (size_t i = 0; i < nx; ++i)
	xx[i] = tao_opt[i] + tao_opt[i+nx];
    }
    else {
      
      // Unconstrained optimization
      
      Unconstrained fn(train, f_c, ccs, p, s);
      tao_optimizer<Unconstrained> tao_opt(nx, fn);

      if (r != 0) 
	for (size_t i = 0; i < nx; ++i)
	  tao_opt[i] = r*double(random()-RAND_MAX/2)/double(RAND_MAX/2);

      // copy the default starting point into tao_opt[]

      for (size_t i = 0; i < nx; ++i)
	tao_opt[i] = x0[i];

      tao_opt.optimize();
      
      nits = fn.it;
      
      for (size_t i = 0; i < nx; ++i)
	xx[i] = tao_opt[i];
    }
    return nits;
  }  // Estimator1::optimize1()


  // operator() actually runs one round of estimation
  //
  double operator() (const doubles& ccs) {
    assert(ccs.size() == nc);
    assert(x.size() == nx);
    assert(x0.size() == nx);
    nits = 0;
    nrounds++;

    /*
    // precompute a good starting point

    if (nrounds == 1) 
      optimize1(ccs, x0);
    */

    if (debug_level >= 10) {
      if (nrounds == 1) 
	std::cout << "# round	ccs	neglogP	f-score" << std::endl;
      std::cout << nrounds << '\t' << ccs << std::flush;
    }

    // Perform the optimization

    nits = optimize1(ccs, x);

    // Clean up, collect stats
    
    sum_nits += nits;

    // Write out weights file
    
    if (!weightsfile.empty()) {
      FILE* out = fopen(weightsfile.c_str(), "w");
      // fprintf(out, "%d@", nx-nzeros);
      for (size_t i = 0; i < x.size(); ++i) 
	if (x[i] != 0) {
	  fprintf(out, "%d", i);
	  if (x[i] != 1)
	    fprintf(out, "=%g", x[i]);
	  fprintf(out, "\n");
	}
      fclose(out);
    }      

    return evaluate(opt_fscore, true);

  }  // Estimator1::operator()

  // evaluate() evaluates the current model on the eval data, prints
  // out debugging information if appropriate, and returns either
  // the - log likelihood or 1 - f-score.
  //
  double evaluate(bool opt_fscore = false, bool internal = false) const {

    std::vector<double> df_dx(nx);
    Float sum_g = 0, sum_p = 0, sum_w = 0;
    Float neglogP = corpus_stats(eval, &x[0], &df_dx[0], 
				 &sum_g, &sum_p, &sum_w);
    Float fscore = 2*sum_w/(sum_g+sum_p);
    
    if (internal) {  // internal evaluation, use a short print-out

      if (debug_level >= 10) 
	std::cout << '\t' << neglogP << '\t' << fscore << std::endl;
    }
    else { // final evaluation, print out more info
    
      int nzeros = 0;
      for (size_t i = 0; i < nx; ++i) 
	if (x[i] == 0)
	  ++nzeros;
  
      std::cout << "# Regularizer power p = " << p << std::endl;
      std::cout << "# " << nx-nzeros << " non-zero feature weights of " 
		<< nx << " features." << std::endl;
      std::cout << "# Eval neglogP = " << neglogP 
		<< ", neglogP/nsentences = " << neglogP/eval->nsentences
		<< std::endl;
      std::cout << "# Eval precision = " << sum_w/sum_p 
		<< ", recall = " << sum_w/sum_g
		<< ", f-score = " << 2*sum_w/(sum_g+sum_p)
		<< std::endl;
      {
	Float sum_g = 0, sum_p = 0, sum_w = 0;
	Float neglogP = corpus_stats(train, &x[0], &df_dx[0], 
				     &sum_g, &sum_p, &sum_w);
	std::cout << "# Test neglogP = " << neglogP 
		  << ", neglogP/nsentences = " << neglogP/train->nsentences
		  << std::endl;
	std::cout << "# Test precision = " << sum_w/sum_p 
		  << ", recall = " << sum_w/sum_g
		  << ", f-score = " << 2*sum_w/(sum_g+sum_p)
		  << std::endl;
      }

      std::cout << "# regclass_identifiers = " << regclass_identifiers << std::endl;
      std::cout << "# cs = " << cs << std::endl;

      if (debug_level >= 100) {
	std::cout << "# Cumulative distribution of feature weights:" << std::endl;
	print_histogram(nx, &x[0]);
      }
    }

    return opt_fscore ? 1 - fscore : neglogP;
  } // Estimator1::evaluate()

  //! fc_bin() maps a feature count to its corresponding bin
  //
  static int fc_bin(double feature_count_base, int feature_count) {
    if (feature_count <= 4)
      return feature_count;
    else
      return lrint(log(feature_count-4)/log(feature_count_base));
  }  // Estimator1::fc_bin()

  //! read_featureclasses() reads the feature classes from a feature file
  //
  void read_featureclasses(const char* filename, 
			   double feature_count_base = 0.0,
			   int nseparators = 1,
			   const char* separators = ":") {
    
    std::string command("bzcat ");
    command += filename;
    FILE* in = popen(command.c_str(), "r");

    if (in == NULL) {
      std::cerr << "## Error: can't popen command " << command << std::endl;
      exit(EXIT_FAILURE);
    }

    size_t featno;
    size_t nf = 0;

    // read feature number 

    while (fscanf(in, " %u ", &featno) == 1) {
      if (featno >= nx) {
	std::cerr << "## Error: read featno = " << featno << ", nx = " << nx << std::endl;
	exit(EXIT_FAILURE);
      }
      ++nf;
      int c = ':';
      
      // read the prefix of the feature class identifier
      
      std::string identifier;
      int iseparators = 0;
      if (nseparators >= 0)
	while ((c = getc(in)) != EOF && !isspace(c)) {
	  if (index(separators, c) != NULL)
	    if (++iseparators > nseparators)
	      break;
	  identifier.push_back(c);
	}
      
      // skip to the next space

      if (c != EOF && !isspace(c))
	while ((c = getc(in)) != EOF && !isspace(c))
	  ;

      // read number
      
      int count = 0;
      int nread = fscanf(in, " %d", &count);

      if (nread == 1 && feature_count_base != 0) 
	identifier = boost::lexical_cast<std::string>(fc_bin(feature_count_base,count))
	  + ":" + identifier;

      // skip the rest of the line

      while ((c = getc(in)) != EOF && c != '\n')
	;

      // insert the prefix into the prefix -> regularization class map
      
      S_C::iterator it 
	= identifier_regclass.insert(S_C::value_type(identifier, 
						     identifier_regclass.size())).first;
      
      size_t cl = it->second;    // regularization class

      assert(featno < f_c.size());
      f_c[featno] = cl;          // set feature's regularization class
    }
      
    if (nx != nf) {
      std::cerr << "## Error: only read " << nf << " features from " 
		<< filename << ", but data contains " << nx << " features."
		<< std::endl;
      exit(EXIT_FAILURE);
    }

    nc = identifier_regclass.size();   // set nc
    cs.resize(nc, c0);                 // set each regularizer class' factor to c0

    // construct regclass_identifiers
    
    regclass_identifiers.resize(nc);
    cforeach (S_C, it, identifier_regclass) {
      assert(it->second < regclass_identifiers.size());
      regclass_identifiers[it->second] = it->first;
    }

    if (debug_level >= 10) 
      std::cout << "# Regularization classes: " << regclass_identifiers 
		<< std::endl;

    pclose(in);
  }  // Estimator1::read_featureclasses()
    
  void estimate()
  {
    powell::control cntrl(1e-4, 1e-2, 0, 5);
    powell::minimize(cs, *this, 1.0, cntrl);
  }  // Estimator1::estimate()

};  // Estimator1{}


int main(int argc, char** argv) 
{
  std::ios::sync_with_stdio(false);
  
  // Initialize TAO and PETSc

  tao_environment tao_env(argc, argv);

  if (tao_env.get_bool_option("-help")) {
    std::cerr << "-help\n" << usage << std::endl;
    exit(EXIT_SUCCESS);
  }

  debug_level = tao_env.get_int_option("-debug", debug_level);
  int ne = tao_env.get_int_option("-ne", 1000);
  int nt = tao_env.get_int_option("-nv", 50000);
  double c0 = tao_env.get_double_option("-c0", 2.0);
  double p = tao_env.get_double_option("-p", 2.0);
  double r = tao_env.get_double_option("-r", 0.0);
  double s = tao_env.get_double_option("-s", 1.0);
  bool cv = tao_env.get_bool_option("-cv");
  double feature_count_base = tao_env.get_double_option("-fcb", 0.0);
  double Pyx_factor = tao_env.get_double_option("-Pyx_factor", 0.0);
  bool Px_propto_g = tao_env.get_bool_option("-Px_propto_g");

  if (debug_level >= 10)
    std::cout << "#  Regularization: c0 = " << c0
	      << ", power p = " << p 
	      << ", scale s = " << s
	      << "; random init r = " << r 
	      << ", constrained var optimization cv = " << cv
	      << ", feature_count_base fcb = " << feature_count_base
	      << ", Pyx_factor = " << Pyx_factor
	      << ", Px_propto_g = " << Px_propto_g
	      << std::endl;

  corpusflags_type corpusflags = { Pyx_factor, Px_propto_g };
  corpus_type* traindata = read_corpus(&corpusflags, stdin, nt);
  int nx = traindata->nfeatures;

  // Read in eval data
  
  corpus_type* evaldata = traindata;
  const char* evalfile = tao_env.get_cstr_option("-e");
  if (evalfile) {
    size_t len = strlen(evalfile);
    bool popen_flg = (len >= 4 && strcmp(evalfile+len-4, ".bz2") == 0);
    FILE *in;
    if (popen_flg) {
      std::string command("bzcat ");
      command += evalfile;
      in = popen(command.c_str(), "r");
    }
    else
      in = fopen(evalfile, "r");
    if (in == NULL) {
      std::cerr << "## Couldn't open evalfile " << evalfile
		<< ", errno = " << errno << "\n" 
		<< usage << std::endl;
      exit(EXIT_FAILURE);
    }
    evaldata = read_corpus(&corpusflags, in, ne);
    if (popen_flg)
      pclose(in);
    else
      fclose(in);
    int nxe = evaldata->nfeatures;
    assert(nxe <= nx);
  }

  Estimator1 e(traindata, evaldata, c0, p, r, s, cv, true,
	       tao_env.get_cstr_option("-o"));

  int nseparators = tao_env.get_int_option("-ns", 1);
  const char* filename = tao_env.get_cstr_option("-f");
  if (filename != NULL)
    e.read_featureclasses(filename, feature_count_base, nseparators, ":");  

  const char* ci = tao_env.get_cstr_option("-ci");
  if (ci != NULL) {
    std::string sci(ci);
    std::istringstream is(sci);
    is >> e.cs;
    if (e.cs.size() != e.nc) {
      std::cerr << "## Error: can't initialize e.cs = " << e.cs 
		<< ", ci = " << ci << std::endl;
      exit(EXIT_FAILURE);
    }
  }

  e.estimate();

}  // main()
