// wlle.cc -- A constrained-variable log-likelihood optimizer
//
// Mark Johnson, 29th July 2003, last modified 26th August 2003
//
// This optimizier uses the TAO constrained optimization method tao_blmvm
// to avoid the variable discontinuity at zero.
//
// For each variable x[i] in the original problem we introduce a pair of
// variables xp[i] and xn[i] and a pair of constraints xp[i] >= 0, xn[i] <= 0.
// Then x[i] = xp[i] + xn[i].
//
// The regularized function Q that is optimized is:
//
//  Q(xp,xn) = s * ( L(xp+xn) + R(xp,xn) )
//
// where s is a user-specified parameter, L is the unregularized loss 
// function and R is the regularizer:
//
//  R(xp,xn) = c * sum_i (pow(fabs(xp[i]),p) + pow(fabs(xn[i]),p))
//
// where c and p are user-specifier adjustable parameters.  (The fabs()
// is there just in case the optimization routine temporarily proposes an
// infeasible solution).
//
// Note that
//
//  d Q/d xp[i] = s * d L/ d xp[i] + s * c * p * sum_i pow(fabs(xp[i]),p-1) * sign(xp[i])
//
// where sign(xp[i]) is +1 if xp[i] is non-negative and -1 otherwise.

const char usage[] =
"wlle version of 26th August 2003\n"
"\n"
"wlle estimates feature weights that estimate the parameters of a\n"
"log-linear model by minimizing the regularized log loss of the feature\n"
"weights using the LVLM or BLVLM optimizer from the Petsc/Tao\n"
"optimization package (see http://www-fp.mcs.anl.gov/tao/ for details).\n"
"It can deal with partially labeled data (i.e., training instances\n"
"consist of one or more \"winners\" and one or more \"losers\").\n"
"\n"
"Usage: wlle [-help] [-debug debug_level] [-c c] [-p p] [-r r] [-s s] [-cv] \n"
"       tao-options* [-o weights-file]  [-e eval-file] < train-file\n"
"\n"
"where:\n"
"\n"
" debug_level > 0 controls the amount of output produced\n"
"\n"
" data respectively (best if slight over-estimates, ignored if the data\n"
" specifies NS),\n"
"\n"
" train-file and eval-file are files from which training and evaluation\n"
" data are read (if no eval-file is specified, then the program tests\n"
" on the training data),\n"
"\n"
" weights-file is a file to which the estimated weights are written,\n"
"\n"
" r specifies that the weights are initialized to random values in\n"
"   [-r ... +r],\n"
"\n"
" and the function optimized is:\n"
"\n"
"   Q(w) = s * (- L(w) + c * sum_j pow(fabs(w[j]), p) ), where\n"
"\n"
"   L(w) = log conditional likelihood of the training data\n"
"\n"
"The -cv option instructs the program to optimize a function defined in\n"
"terms of vectors of variables u[] and v[], where w[j] = u[j] + v[j]\n"
"and v[j] <= 0 <= u[j], otherwise it optimizes the w[j] directly.\n"
"\n"
"With debug = 0, the program writes a single line to stdout:\n"
"\n"
"c p r s it nzeroweights/nweights neglogP/nsentence ncorrect/nsentences\n"
"\n"
"With debug >= 10, the program writes out a histogram of weights as well\n"
"\n"
"Data format:\n"
"-----------\n"
"\n"
"<Data>     --> [S=<NS>] <Sentence>*\n"
"<Sentence> --> [C=<C>] N=<N> <Parse>*\n"
"<Parse>    --> [W=<W>] <FC>*,\n"
"<FC>       --> <F>[=<C>]\n"
"\n"
"NS is the number of sentences.\n"
"\n"
"Each <Sentence> consists of N <Parse>s.  <C> is the sentence's count\n"
"(a real, e.g., the number of times it was seen).\n"
"\n"
"A <Parse> consists of <FC> pairs.  <W> is the parse's weight.  Winners\n"
"should have a weight of 1.\n"
"\n"
"A <FC> consists of a feature (a non-negative integer) and a count (a\n"
"real).\n"
"\n"
"The default for all numbers except <W> is 1.  The default for <W> is 0.\n";

#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <vector>

#include "data.h"
#include "tao.h"

int debug_level = 0;

void print_histogram(int nx, double x[]) {
  int nx_nonzero = 0;
  for (int i = 0; i < nx; ++i)
    if (x[i] != 0)
      ++nx_nonzero;

  std::cout << "#   There are " << nx_nonzero << " non-zero values and " 
	    << nx-nx_nonzero << " zero values." << std::endl;

  if (nx_nonzero > 0) {
    std::vector<double> s;
    s.reserve(nx_nonzero);
    for (int i = 0; i < nx; ++i)
      if (x[i] != 0)
	s.push_back(x[i]);
    std::sort(s.begin(), s.end());
    for (int i = 0; i <= 20; ++i) {
      int j = i*(nx_nonzero-1);
      j /= 20;
      std::cout << i/20.0 << '\t' << s[j] << std::endl;
    }
  }
}  // print_histogram()


double p_norm(double p, int nx, double x[]) { // returns p-norm of x
  double s = 0;
  for (int i = 0; i < nx; ++i)
    s += pow(fabs(x[i]), p);
  return (s == 0) ? 0 : pow(s, 1.0/p);
}  // p_norm


// divide_by_p_norm() divides each y[i] by ||x||_p, returns ||x||_p
//
double divide_by_p_norm(double p, int ny, double y[], int nx, double x[])
{
  double pnorm = p_norm(p, nx, x);
  assert(pnorm > 0);
  for (int j = 0; j < ny; ++j)
    y[j] /= pnorm;
  return pnorm;
}  // divide_by_pnorm()


// AppCntx{} and callback() are for unconstrained optimization
//
struct AppCntx {
  corpus_type* corpus;
  size_t nx;
  double c, p, s;
  int it;

  AppCntx(corpus_type* corpus, double c, double p, double s) 
    : corpus(corpus), nx(corpus->nfeatures), c(c), p(p), s(s), it(0) { }

  double operator() (int nx, double x[], double df_dx[]) {
    it++;
    if (debug_level >= 1000)      std::cout << "it = " << it << ", " << std::flush;

    Float ncorrect;
    Float L = corpus_stats(corpus, &x[0], &df_dx[0], &ncorrect);
    if (debug_level >= 1000)
      std::cout << "ncorrect = " << ncorrect << ", " << std::flush;

    assert(finite(L));
    
    if (s != 1) {
      L *= s;

      for (int i = 0; i < nx; ++i) {
	assert(finite(df_dx[i]));
	df_dx[i] *= s;
      }
    }

    if (c == 0) {
      if (debug_level >= 1000)
	std::cout << std::endl;
      return L;
    }

    double R = 0;
    for (int i = 0; i < nx; ++i)
      R += pow(fabs(x[i]), p);
    R *= s * c;

    double Q = L + R;
    
    if (debug_level >= 1000)
      std::cout << "Q = " << Q << " = L = " << L << " + R = " << R << std::endl;

    assert(finite(Q));
    
    double scp = s * c * p;
    for (int i = 0; i < nx; ++i) 
      df_dx[i] += scp * pow(fabs(x[i]), p-1) 
	* (x[i] >= 0 ? (x[i] == 0 ? 0 : 1) : -1);

    if (debug_level >= 10000) {
      std::cout << "Histogram of derivatives:" << std::endl;
      print_histogram(nx, &df_dx[0]);
      std::cout << "--------------------------" << std::endl;
    }

    for (int i = 0; i < nx; ++i)
      assert(finite(df_dx[i]));
    return Q;
  }   // AppCntx::operator()

};  // AppCntx{}

int callback(TAO_APPLICATION taoapp, Vec X, double *fx, 
	     Vec DF_DX, void *userCtx)
{
  typedef AppCntx* AppCntxPtr;
  AppCntx& appcntx = *AppCntxPtr(userCtx);

  double* x;
  int info;
  info = VecGetArray(X, &x); assert(info == 0);
  double* df_dx;
  info = VecGetArray(DF_DX, &df_dx); assert(info == 0);

  int nx;
  info = VecGetSize(X, &nx); assert(info == 0);
  int ndf_dx;
  info = VecGetSize(DF_DX, &ndf_dx); assert(info == 0);
  assert(nx == ndf_dx);
   
  // Compute f(x) and df_dx
   
  *fx = appcntx(nx, x, df_dx);

  // Restore vectors
   
  info = VecRestoreArray(X, &x); assert(info == 0);
  info = VecRestoreArray(DF_DX, &df_dx); assert(info == 0);
   
  return 0;
}  // callback()


// AppCntx2{} and callback2() are for the constrained optimization.
//
struct AppCntx2 {
  corpus_type* corpus;
  size_t nx;
  double c, p, s;
  int it;
  std::vector<double> x, df_dx;

  AppCntx2(corpus_type* corpus, double c, double p, double s) 
    : corpus(corpus), nx(corpus->nfeatures), c(c), p(p), s(s), it(0), 
      x(nx), df_dx(nx) { }

  double operator() (int nx2, double x2[], double df_dx2[]) {
    it++;
    if (debug_level >= 1000)      std::cout << "it = " << it << ", " << std::flush;

    assert(nx2 = 2 * nx);
    assert(x.size() == nx);
    for (size_t i = 0; i < nx; ++i) {
      assert(finite(x2[i]));
      assert(finite(x2[i+nx]));
      x[i] = x2[i] + x2[i+nx];
      assert(finite(x[i]));
    }

    Float ncorrect;
    Float L = corpus_stats(corpus, &x[0], &df_dx[0], &ncorrect);
    if (debug_level >= 1000)
      std::cout << "ncorrect = " << ncorrect << ", " << std::flush;

    L *= s;
    assert(finite(L));

    for (size_t i = 0; i < nx; ++i) {
      assert(finite(df_dx[i]));
      df_dx2[i] = df_dx2[i+nx] = s * df_dx[i];
    }

    if (c == 0)
      return L;

    double R = 0;
    for (size_t i = 0; i < nx; ++i)
      R += pow(fabs(x2[i]), p) + pow(fabs(x2[i+nx]), p);
    R *= s * c;

    double Q = L + R;
    
    if (debug_level >= 1000)
      std::cout << "Q = " << Q << " = L = " << L << " + R = " << R << std::endl;

    assert(finite(Q));
    
    double scp = s * c * p;
    for (size_t i = 0; i < nx; ++i) {
      df_dx2[i] += scp * pow(fabs(x2[i]), p-1) * (x2[i] >= 0 ? 1 : -1);
      df_dx2[i+nx] += scp * pow(fabs(x2[i+nx]), p-1) * (x2[i+nx] > 0 ? 1 : -1);
    }

    if (debug_level >= 10000) {
      std::cout << "Histogram of derivatives:" << std::endl;
      print_histogram(nx, &df_dx[0]);
      std::cout << "--------------------------" << std::endl;
    }

    for (int i = 0; i < nx2; ++i)
      assert(finite(df_dx2[i]));
    return Q;
  }   // AppCntx2::operator()

};  // AppCntx2

int callback2(TAO_APPLICATION taoapp, Vec X2, double *fx, 
	      Vec DF_DX2, void *userCtx)
{
  typedef AppCntx2* AppCntxPtr;
  AppCntx2& appcntx = *AppCntxPtr(userCtx);

  double* x2;
  int info;
  info = VecGetArray(X2, &x2); assert(info == 0);
  double* df_dx2;
  info = VecGetArray(DF_DX2, &df_dx2); assert(info == 0);

  int nx2;
  info = VecGetSize(X2, &nx2); assert(info == 0);
  int ndf_dx2;
  info = VecGetSize(DF_DX2, &ndf_dx2); assert(info == 0);
  assert(nx2 == ndf_dx2);
   
  // Compute f(x) and df_dx
   
  *fx = appcntx(nx2, x2, df_dx2);

  // Restore vectors
   
  info = VecRestoreArray(X2, &x2); assert(info == 0);
  info = VecRestoreArray(DF_DX2, &df_dx2); assert(info == 0);
   
  return 0;
}  // callback2()


int main(int argc, char** argv) 
{
  // Initialize TAO and PETSc
  
  int info;
  info = TaoInitialize(&argc, &argv, (char *) 0, usage); assert(info == 0);
  
  // Start optimization
  
  PetscTruth help = PETSC_FALSE;
  info = PetscOptionsGetLogical(PETSC_NULL, "-help", &help, PETSC_NULL); assert(info == 0);
  if (help) {
    std::cerr << usage << std::endl;
    exit(EXIT_SUCCESS);
  }

  char filename[512];
  PetscTruth flg = PETSC_FALSE;
  info = PetscOptionsGetString(PETSC_NULL, "-e", filename, 511, &flg); assert(info == 0);
  
  info = PetscOptionsGetInt(PETSC_NULL, "-debug", &debug_level, PETSC_NULL); assert(info == 0);

  int ne = 1000;
  info = PetscOptionsGetInt(PETSC_NULL, "-ne", &ne, PETSC_NULL); assert(info == 0);

  int nt = 50000;
  info = PetscOptionsGetInt(PETSC_NULL, "-nt", &nt, PETSC_NULL); assert(info == 0);

  double c = 0.0;
  info = PetscOptionsGetReal(PETSC_NULL, "-c", &c, PETSC_NULL); assert(info == 0);

  double p = 2.0;
  info = PetscOptionsGetReal(PETSC_NULL, "-p", &p, PETSC_NULL); assert(info == 0);
  
  double r = 0.0;
  info = PetscOptionsGetReal(PETSC_NULL, "-r", &r, PETSC_NULL); assert(info == 0);

  double s = 1.0;
  info = PetscOptionsGetReal(PETSC_NULL, "-s", &s, PETSC_NULL); assert(info == 0);

  PetscTruth cv = PETSC_FALSE;
  info = PetscOptionsGetLogical(PETSC_NULL, "-cv", &cv, PETSC_NULL); assert(info == 0);

  if (debug_level > 10)
    std::cout << "#  Regularization: c = " << c 
	      << ", power p = " << p 
	      << ", scale s = " << s
	      << "; random init r = " << r 
	      << ", constrained var optimization cv = " << cv
	      << std::endl;

  corpus_type* traindata = read_corpus(stdin, nt);
  int nx = traindata->nfeatures;

  corpus_type* evaldata = traindata;
  if (flg) {
    FILE* in = fopen(filename, "r");
    if (in == NULL) {
      std::cerr << "## Couldn't open evalfile " << filename
		<< ", errno = " << errno << "\n" 
		<< usage << std::endl;
      exit(EXIT_FAILURE);
    }
    evaldata = read_corpus(in, ne);
    fclose(in);
    int nxe = evaldata->nfeatures;
    assert(nxe <= nx);
  }

  std::vector<double> x;
  int nits = 0;

  if (cv) {
    // Constrained variable optimization
    // Allocate vectors
    
    int nx2 = 2 * nx;  
    Vec X2, X2L, X2U;
    info = VecCreateSeq(PETSC_COMM_SELF, nx2, &X2); assert(info == 0);
    info = VecCreateSeq(PETSC_COMM_SELF, nx2, &X2L); assert(info == 0);
    info = VecCreateSeq(PETSC_COMM_SELF, nx2, &X2U); assert(info == 0);

    if (r == 0) {
      double zero = 0.0;
      info = VecSet(&zero, X2); assert(info == 0);
    }
    else {
      double* x2;
      info = VecGetArray(X2, &x2); assert(info == 0);
      for (int i = 0; i < nx2; ++i)
	x2[i] = r*double(random()-RAND_MAX/2)/double(RAND_MAX/2);
      info = VecRestoreArray(X2, &x2); assert(info == 0);
    }

    {
      double *x2l, *x2u;
      info = VecGetArray(X2L, &x2l); assert(info == 0);
      info = VecGetArray(X2U, &x2u); assert(info == 0);
      for (int i = 0; i < nx; ++i) {
	x2l[i] = 0;
	x2l[i+nx] = TAO_NINFINITY;
	x2u[i] = TAO_INFINITY;
	x2u[i+nx] = 0;
      }
      info = VecRestoreArray(X2L, &x2l); assert(info == 0);
      info = VecRestoreArray(X2U, &x2u); assert(info == 0);
    }
    
    // Create TAO solver
    
    TAO_SOLVER      tao;
    info = TaoCreate(PETSC_COMM_SELF, "tao_blmvm", &tao); assert(info == 0);
    assert(info == 0);
    TAO_APPLICATION tao_application;
    info = TaoApplicationCreate(PETSC_COMM_SELF, &tao_application);
    assert(info == 0);

    TaoAppSetVariableBounds(tao_application, X2L, X2U);

    info = TaoSetFromOptions(tao); assert(info == 0);
    info = TaoSetMethodFromOptions(tao); assert(info == 0);
  
    info = TaoAppSetInitialSolutionVec(tao_application, X2); assert(info == 0);
  
    // Set routine for function evaluation and gradient
    { 
      AppCntx2 cntx(traindata, c, p, s);
      info = TaoAppSetObjectiveAndGradientRoutine(tao_application,
						callback2, (void *) (&cntx));
      assert(info == 0);
      
      // Run the solver
      
      info = TaoSolveApplication(tao_application, tao); assert(info == 0);
      nits = cntx.it;
    }
    // Load the solution into x2
    
    double* x2;
    info = VecGetArray(X2, &x2); assert(info == 0);
    
    x.resize(nx);

    for (int i = 0; i < nx; ++i)
      x[i] = x2[i] + x2[i+nx];

    info = VecDestroy(X2); assert(info == 0);
    info = VecDestroy(X2L); assert(info == 0);
    info = VecDestroy(X2U); assert(info == 0);
  
    info = TaoDestroy(tao); assert(info == 0);
    info = TaoApplicationDestroy(tao_application); assert(info == 0);
  }
  else {
    // Unconstrained optimization
    // Allocate vectors
    
    Vec X;
    info = VecCreateSeq(PETSC_COMM_SELF, nx, &X); assert(info == 0);

    if (r == 0) {
      double zero = 0.0;
      info = VecSet(&zero, X); assert(info == 0);
    }
    else {
      double* x;
      info = VecGetArray(X, &x); assert(info == 0);
      for (int i = 0; i < nx; ++i)
	x[i] = r*double(random()-RAND_MAX/2)/double(RAND_MAX/2);
      info = VecRestoreArray(X, &x); assert(info == 0);
    }

    // Create TAO solver
    
    TAO_SOLVER      tao;
    info = TaoCreate(PETSC_COMM_SELF, "tao_lmvm", &tao); assert(info == 0);
    assert(info == 0);
    TAO_APPLICATION tao_application;
    info = TaoApplicationCreate(PETSC_COMM_SELF, &tao_application);
    assert(info == 0);

    info = TaoSetFromOptions(tao); assert(info == 0);
    info = TaoSetMethodFromOptions(tao); assert(info == 0);
  
    info = TaoAppSetInitialSolutionVec(tao_application, X); assert(info == 0);
  
    // Set routine for function evaluation and gradient
    { 
      AppCntx cntx(traindata, c, p, s);
      info = TaoAppSetObjectiveAndGradientRoutine(tao_application,
						  callback, (void *) (&cntx));
      assert(info == 0);
      
      // Run the solver
      
      info = TaoSolveApplication(tao_application, tao); assert(info == 0);
      nits = cntx.it;
    }
    // Load the solution into x
    
    double* xx;
    info = VecGetArray(X, &xx); assert(info == 0);
    
    x.resize(nx);

    for (int i = 0; i < nx; ++i)
      x[i] = xx[i];

    info = VecDestroy(X); assert(info == 0);
  
    info = TaoDestroy(tao); assert(info == 0);
    info = TaoApplicationDestroy(tao_application); assert(info == 0);
  }
   
  int nzeros = 0;
  for (int i = 0; i < nx; ++i) 
    if (x[i] == 0)
      ++nzeros;
  
  std::vector<double> df_dx(nx);
  Float ncorrect, neglogP = corpus_stats(evaldata, &x[0], &df_dx[0], &ncorrect);
  if (debug_level > 0)
    std::cout << "# c p r s it nzeroweights/nweights neglogP/nsentences ncorrect/nsentences" << std::endl;
  std::cout << c << '\t' << p << '\t' << r << '\t' << s 
	    << '\t' << nits << '\t' << double(nzeros)/nx 
	    << '\t' << neglogP/evaldata->sum_count
	    << '\t' << ncorrect/evaldata->sum_count << std::endl;
  
  if (debug_level >= 10) {
    std::cout << "# Cumulative distribution of feature weights:" << std::endl;
    print_histogram(nx, &x[0]);
  }
    
  info = PetscOptionsGetString(PETSC_NULL, "-o", filename, 511, &flg); assert(info == 0);
  if (flg) {
    FILE* out = fopen(filename, "w");
    fprintf(out, "%d@", nx-nzeros);
    for (int i = 0; i < nx; ++i) 
      if (x[i] != 0) {
	fprintf(out, " %d", i);
	if (x[i] != 1)
	  fprintf(out, "=%g", x[i]);
      }
    fclose(out);
  }

  TaoFinalize();
}  // main()
