// lnne.cc -- A logistic neural net estimator
//
// Mark Johnson, 7th October 2003
//
// This is a regularized logistic neural net estimator using TAO.
//
//
// The regularized function Q that is optimized is:
//
//  Q(xp,xn) = s * ( L(xp+xn) + R(xp,xn) )
//
// where s is a user-specified parameter, L is the unregularized loss 
// function and R is the regularizer:
//
//  R(xp,xn) = c * sum_i (pow(fabs(xp[i]),p) + pow(fabs(xn[i]),p))
//
// where c and p are user-specifier adjustable parameters.  (The fabs()
// is there just in case the optimization routine temporarily proposes an
// infeasible solution).
//
// Note that
//
//  d Q/d xp[i] = s * d L/ d xp[i] + s * c * p * sum_i pow(fabs(xp[i]),p-1) * sign(xp[i])
//
// where sign(xp[i]) is +1 if xp[i] is non-negative and -1 otherwise.

const char usage[] =
"lnne version of 7th October 2003\n"
"\n"
"llne estimates the parameters of a logistic neural network using\n"
"the LVLM optimizer from the Petsc/Tao optimization package\n"
"(see http://www-fp.mcs.anl.gov/tao/ for details).\n"
"It can deal with partially labeled data (i.e., training instances\n"
"consist of one or more \"winners\" and one or more \"losers\").\n"
"\n"
"Usage: llne [-help] [-debug debug_level] [-nhidden nh] [-c c] [-p p] [-r r] [-s s] [-cv] \n"
"            [-Pyx_factor f] [-Px_propto_g] \n"
"            tao-options* [-o weights-file]  [-e eval-file] [-f feat-file] < train-file\n"
"\n"
"where:\n"
"\n"
" debug_level > 0 controls the amount of output produced\n"
"\n"
" data respectively (best if slight over-estimates, ignored if the data\n"
" specifies NS),\n"
"\n"
" train-file and eval-file are files from which training and evaluation\n"
" data are read (if eval-file ends in the suffix .bz2 then bzcat is used\n"
" to read it; if no eval-file is specified, then the program tests on the\n"
" training data),\n"
"\n"
" weights-file is a file to which the estimated weights are written,\n"
"\n"
" feat-file is a file of <featclass> <featuredetails> lines, used for\n"
" cross-validating regularizer weights,\n"
"\n"
" nh is the number of hidden units,\n"
"\n"
" r specifies that the weights are initialized to random values in\n"
"   [-r ... +r],\n"
"\n"
" -Pyx_factor f indicates that a parse should be taken as correct\n"
"   proportional to f raised to its f-score, and\n"
"\n"
" -Px_propto_g indicates that each sentence is weighted by the number of\n"
"   edges in its gold parse.\n"
"\n"
"With debug = 0, the program writes a single line to stdout:\n"
"\n"
"c p r s it nzeroweights/nweights neglogP/nsentence ncorrect/nsentences\n"
"\n"
"\n"
"Data format:\n"
"-----------\n"
"\n"
"<Data>     --> [S=<NS>] <Sentence>*\n"
"<Sentence> --> [G=<G>] N=<N> <Parse>*\n"
"<Parse>    --> [P=<P>] [W=<W>] <FC>*,\n"
"<FC>       --> <F>[=<C>]\n"
"\n"
"NS is the number of sentences.\n"
"\n"
"Each <Sentence> consists of N <Parse>s.  <G> is the gold standard\n"
"score.  To get parsing precision and recall results, set <G> to the\n"
"number of edges in the gold standard parse.  To get accuracy results,\n"
"set <G> to 1 (the default).\n"
"\n"
"A <Parse> consists of <FC> pairs.  <P> is the parse's possible highest\n"
"score and <W> is the parse's actual score.  To get parsing precision and\n"
"recall results, set <P> to the number of edges in the parse and <W> to\n"
"the number of edges in common between the gold and parse trees.\n"
"\n"
"A <FC> consists of a feature (a non-negative integer) and an optional\n"
"count (a real).\n"
"\n"
"The default for all numbers except <W> is 1.  The default for <W> is 0.\n";

#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <vector>

#include "data.h"
#include "tao-optimizer.h"

int debug_level = 0;

// f_df() evaluates the statistics of the corpus, and prints the f-score
//  if required
//
double f_df(corpus_type* corpus, size_t nhidden, double x[], double df_dx[]) {
  Float sum_g = 0, sum_p = 0, sum_w = 0;
  Float L = lnn_corpus_stats(corpus, nhidden, &x[0], &df_dx[0], &sum_g, &sum_p, &sum_w);
  
  if (debug_level >= 1000)
    std::cout << "f score = " << 2*sum_w/(sum_g+sum_p) << ", " << std::flush;
  
  assert(finite(L));
  return L;
}


// Unconstrained is for unconstrained optimization
//
struct Unconstrained {
  corpus_type* corpus;
  size_t nhidden;
  size_t nx;
  double c, p, s;
  int it;
  corpus_type* eval_corpus;

  Unconstrained(corpus_type* corpus, size_t nhidden, double c, double p, double s, 
		corpus_type* eval_corpus = NULL) 
    : corpus(corpus), nhidden(nhidden), nx(nhidden*(corpus->nfeatures+2)), 
      c(c), p(p), s(s), it(0), eval_corpus(eval_corpus) { }

  double operator() (int nx, double x[], double df_dx[]) {
    it++;
    if (debug_level >= 1000)      std::cout << "it = " << it << ", " << std::flush;

    assert(size_t(nx) == nhidden*(corpus->nfeatures+2));

    if (debug_level >= 2000 && eval_corpus != NULL) {
      Float sum_g = 0, sum_p = 0, sum_w = 0;
      std::cout << "(Eval corpus " << std::flush;
      Float L = lnn_corpus_stats(eval_corpus, nhidden, &x[0], &df_dx[0], &sum_g, &sum_p, &sum_w);
      std::cout << "L = " << L << ", f score = " << 2*sum_w/(sum_g+sum_p) << ") " << std::flush;
    }

    Float L = f_df(corpus, nhidden, x, df_dx);
    
    if (s != 1) {
      L *= s;

      for (int i = 0; i < nx; ++i) {
	assert(finite(df_dx[i]));
	df_dx[i] *= s;
      }
    }

    if (c == 0) {
      if (debug_level >= 1000)
	std::cout << std::endl;
      return L;
    }

    double R = 0;
    for (int i = 0; i < nx; ++i)
      R += pow(fabs(x[i]), p);
    R *= s * c;

    double Q = L + R;
    
    if (debug_level >= 1000)
      std::cout << "Q = " << Q << " = L = " << L << " + R = " << R << std::endl;

    assert(finite(Q));
    
    double scp = s * c * p;
    for (int i = 0; i < nx; ++i) {
      df_dx[i] += scp * pow(fabs(x[i]), p-1) 
	* (x[i] >= 0 ? (x[i] == 0 ? 0 : 1) : -1);
      assert(finite(df_dx[i]));
    }

    return Q;
  }   // Unconstrained::operator()

};  // Unconstrained{}



int main(int argc, char** argv) 
{
  std::ios::sync_with_stdio(false);
  
  // Initialize TAO and PETSc

  tao_environment tao_env(argc, argv);

  if (tao_env.get_bool_option("-help")) {
    std::cerr << "-help\n" << usage << std::endl;
    exit(EXIT_SUCCESS);
  }

  debug_level = tao_env.get_int_option("-debug", debug_level);
  int nhidden = tao_env.get_int_option("-nhidden", 5);

  int ne = tao_env.get_int_option("-ne", 1000);
  int nt = tao_env.get_int_option("-nv", 50000);
  double c = tao_env.get_double_option("-c", 0.0);
  double p = tao_env.get_double_option("-p", 2.0);
  double r = tao_env.get_double_option("-r", 0.0);
  double s = tao_env.get_double_option("-s", 1.0);
  double Pyx_factor = tao_env.get_double_option("-Pyx_factor", 0.0);
  bool Px_propto_g = tao_env.get_bool_option("-Px_propto_g");

  if (debug_level >= 10)
    std::cout << "# nhidden = " << nhidden
	      << ", regularization: c = " << c 
	      << ", power p = " << p 
	      << ", scale s = " << s
	      << "; random init r = " << r 
	      << ", Pyx_factor = " << Pyx_factor
	      << ", Px_propto_g = " << Px_propto_g
	      << std::endl;

  corpusflags_type corpusflags = { Pyx_factor, Px_propto_g };
  corpus_type* traindata = read_corpus(&corpusflags, stdin, nt);
  int nx = traindata->nfeatures;

  // Read in eval data
  
  corpus_type* evaldata = traindata;
  const char* evalfile = tao_env.get_cstr_option("-e");
  if (evalfile) {
    size_t len = strlen(evalfile);
    bool popen_flg = (len >= 4 && strcmp(evalfile+len-4, ".bz2") == 0);
    FILE *in;
    if (popen_flg) {
      std::string command("bzcat ");
      command += evalfile;
      in = popen(command.c_str(), "r");
    }
    else
      in = fopen(evalfile, "r");
    if (in == NULL) {
      std::cerr << "## Couldn't open evalfile " << evalfile
		<< ", errno = " << errno << "\n" 
		<< usage << std::endl;
      exit(EXIT_FAILURE);
    }
    evaldata = read_corpus(&corpusflags, in, ne);
    if (popen_flg)
      pclose(in);
    else
      fclose(in);
    int nxe = evaldata->nfeatures;
    assert(nxe <= nx);
  }

  std::vector<double> x;
  int nits = 0;

  {

    // Unconstrained optimization

    Unconstrained fn(traindata, nhidden, c, p, s, evaldata);
    tao_optimizer<Unconstrained> tao_opt(nhidden*(nx+2), fn);

    if (r != 0) 
      for (int i = 0; i < nx; ++i)
	tao_opt[i] = r*double(random()-RAND_MAX/2)/double(RAND_MAX/2);

    tao_opt.optimize();

    nits = fn.it;
    x.resize(nhidden*(nx+2));

    for (int i = 0; i < nhidden*(nx+2); ++i)
      x[i] = tao_opt[i];
  }
   
  std::vector<double> df_dx(nhidden*(nx+2));
  Float sum_g = 0, sum_p = 0, sum_w = 0;
  Float neglogP = lnn_corpus_stats(evaldata, nhidden, &x[0], &df_dx[0], &sum_g, &sum_p, &sum_w);
  if (debug_level > 0)
    std::cout << "# c p r s it neglogP/nsentences precision recall f-score" << std::endl;
  std::cout << c << '\t' << p << '\t' << r << '\t' << s 
	    << '\t' << nits 
	    << '\t' << neglogP/evaldata->nsentences
	    << '\t' << sum_w/sum_p 
	    << '\t' << sum_w/sum_g
	    << '\t' << 2*sum_w/(sum_g+sum_p)
	    << std::endl;
  
  if (debug_level >= 10) {
    sum_g = 0; sum_p = 0; sum_w = 0;
    neglogP = lnn_corpus_stats(traindata, nhidden, &x[0], &df_dx[0], &sum_g, &sum_p, &sum_w);
    std::cout << "# Training log likelihood / nsentences = " 
	      << neglogP/traindata->nsentences << std::endl;
  }

  const char* filename = tao_env.get_cstr_option("-o");

  if (filename != NULL) {
    FILE* out = fopen(filename, "w");
    lnn_weights_type wt;
    lnn_unpack_weights_type(&x[0], nhidden, nx, &wt);
    
    for (int j = 0; j < nhidden; ++j) 
      fprintf(out, "%d\t%g\t%g\n", j, wt.w1[j], wt.b0[j]);
      
    for (int j = 0; j < nx; ++j) {
      fprintf(out, "%d", j);
      for (int k = 0; k < nhidden; ++k)
	fprintf(out, "\t%g", wt.w0[k*nx+j]);
      fprintf(out, "\n");
    }
  }
}  // main()
